/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/fs/fs.hpp>

#include "assetmanager.hpp"
#include "event.hpp"

namespace nostalgia::common {
class Size;
}

namespace nostalgia::core {

class Context;

class Drawer {
	public:
		virtual ~Drawer() = default;
		virtual void draw(Context*) noexcept = 0;
};

namespace renderer {
ox::Error init(Context *ctx) noexcept;
ox::Error shutdown(Context *ctx) noexcept;
ox::Error loadBgTexture(Context *ctx, int section, void *pixels, int w, int h) noexcept;
}

// User Input Output
class Context {
	friend bool bgStatus(Context *ctx, unsigned bg) noexcept;
	friend common::Size getScreenSize(Context *ctx) noexcept;
	friend int getScreenHeight(Context *ctx) noexcept;
	friend int getScreenWidth(Context *ctx) noexcept;
	friend ox::Error initGfx(Context *ctx) noexcept;
	friend ox::Error renderer::init(Context *ctx) noexcept;
	friend ox::Error renderer::loadBgTexture(Context *ctx, int section, void *pixels, int w, int h) noexcept;
	friend ox::Error renderer::shutdown(Context *ctx) noexcept;
	friend ox::Error run(Context *ctx) noexcept;
	friend ox::Error shutdown(Context *ctx) noexcept;
	friend ox::Result<ox::UniquePtr<Context>> init(ox::UniquePtr<ox::FileSystem> fs, const char *appName) noexcept;
	friend ox::String getClipboardText(Context *ctx) noexcept;
	friend uint64_t ticksMs(Context *ctx) noexcept;
	friend uint8_t bgStatus(Context *ctx) noexcept;
	friend void clearTileLayer(Context *ctx, int layer) noexcept;
	friend void draw(Context *ctx) noexcept;
	friend void focusWindow(Context *ctx) noexcept;
	friend void setBgStatus(Context *ctx, uint32_t status) noexcept;
	friend void setBgStatus(Context *ctx, unsigned bg, bool status) noexcept;
	friend void setClipboardText(Context *ctx, const ox::String &text) noexcept;
	friend void setEventHandler(Context *ctx, event_handler h) noexcept;
	friend void setRendererData(Context *ctx, void *rendererData) noexcept;
	friend void setTile(Context *ctx, int layer, int column, int row, uint8_t tile) noexcept;
	friend void setWindowTitle(Context *ctx, const char *title) noexcept;
	friend void setWindowerData(Context *ctx, void *windowerData) noexcept;

	public:
		ox::UniquePtr<ox::FileSystem> rom;
		ox::Vector<Drawer*, 5> drawers;
		const char *appName = "Nostalgia";
#ifndef OX_BARE_METAL
		AssetManager assetManager;
#endif
	private:
		void *m_customData = nullptr;
		void *m_windowerData = nullptr;
		void *m_rendererData = nullptr;

	public:
		Context() noexcept = default;

		Context(Context &other) noexcept = delete;
		Context(const Context &other) noexcept = delete;
		Context(const Context &&other) noexcept = delete;

		constexpr void setCustomData(void *customData) noexcept {
			m_customData = customData;
		}

		template<typename T>
		[[nodiscard]]
		constexpr T *customData() noexcept {
			return static_cast<T*>(m_customData);
		}

	protected:
		template<typename T>
		[[nodiscard]]
		constexpr T *windowerData() noexcept {
			return static_cast<T*>(m_windowerData);
		}

		constexpr void setWindowerData(void *windowerData) noexcept {
			m_windowerData = windowerData;
		}

		constexpr void setRendererData(void *rendererData) noexcept {
			m_rendererData = rendererData;
		}

		template<typename T>
		[[nodiscard]]
		constexpr T *rendererData() noexcept {
			return static_cast<T*>(m_rendererData);
		}

};

}

